import varpepdb.classes as vc
import varpepdb.core as vco
from rpg.enzyme import Enzyme
from varpepdb.core import generate, generate_single, \
    variant_containing_peptides, write, separate_nonunique


def setmiscleave(miscleave: bool) -> None:
    """Turns miscleavage on or off

    Args:
        miscleave: True or False
    """
    vco.miscleave = miscleave


def setenzyme(enzyme_obj: Enzyme) -> None:
    """Select enzyme(s) for cleavage

    Args:
        enzyme_obj: Single or list of enzymes from the ``rpg`` package
    """
    if not isinstance(enzyme_obj, list):
        enzyme_obj = [enzyme_obj]
    vc.Cleaver.enzyme = enzyme_obj
    vc.Variant.enzyme = enzyme_obj


def setpeptidelengths(min_length: int, max_length: int) -> None:
    """Set peptide length limits

    Args:
        min_length
        max_length
    """
    vc.Peptide.min_length = min_length
    vc.Peptide.max_length = max_length


def setdiscardcanonical(discardcanonical: bool) -> None:
    """Discard canonical peptides or not

    Args:
        discardcanonical: True or False
    """
    vco.discardcanonical = discardcanonical


def setcombinations(combinations: bool) -> None:
    """To create combinations of SAPs or not

    Args:
        combinations: True or False
    """
    vco.createcombi = combinations


def setmessage(message: bool) -> None:
    """To turn message off or on. Default is on.

    Args:
        message: True or False
    """
    vco.message = message


AA_1to3 = {'A': 'Ala', 'C': 'Cys', 'D': 'Asp',
           'E': 'Glu', 'F': 'Phe', 'G': 'Gly',
           'H': 'His', 'I': 'Ile', 'K': 'Lys',
           'L': 'Leu', 'M': 'Met', 'N': 'Asn',
           'P': 'Pro', 'Q': 'Gln', 'R': 'Arg',
           'S': 'Ser', 'T': 'Thr', 'V': 'Val',
           'W': 'Trp', 'Y': 'Tyr'}

AA_3to1 = {v: k for k, v in AA_1to3.items()}


def write_fasta(sequence, id, description, fh):
    fh.write(">" + id + ' ' + description + "\n")
    fh.write(sequence + "\n")



import rpg
import rpg.RapidPeptidesGenerator
import varpepdb.classes as vc
import varpepdb
import varpepdb.core as vco
import unittest
import Bio.SeqUtils
import pdb

enzyme_obj = rpg.RapidPeptidesGenerator.ALL_ENZYMES[26]


def setUpModule():
    varpepdb.setenzyme(enzyme_obj)


def generatevariant(peptide, pos, target):
    converter = Bio.SeqUtils.IUPACData.protein_letters_1to3
    target_3 = converter[target]
    source = converter[str(peptide)[pos]]
    start = peptide.start
    return f"{peptide.identifier}:p.{source}{start+pos+1}{target_3}"


class TestGenerateBreakerPeptides(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.peptide = vc.Peptide("FNGIYADPSGHCNFFWPSLFSHYKALMPAGNCI", 'identifier', 'gene')

    def test_novariant(self):
        variants = vc.Variant([], enzyme_obj)
        testres = vco._generate_breaker_peptides(variants=variants, peptide=self.peptide)
        self.assertEqual(str(testres[0]), 'FNGIYADPSGHCNFFW')
        self.assertEqual(str(testres[1]), 'PSLFSHYKALMPAGNCI')

    def test_exclude_enzymevariantsite(self):
        variants = vc.Variant([generatevariant(self.peptide, 15, 'K')], enzyme_obj)
        testres = vco._generate_breaker_peptides(variants=variants, peptide=self.peptide)
        self.assertEqual(str(testres[0]), 'FNGIYADPSGHCNFFWPSLFSHYKALMPAGNCI')


class TestAllocateVariantstoPeptide(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.peptide = vc.Peptide("FNGIYADPSGHCNFFWPSLFSHYKALMPAGNCI", 'identifier', 'gene')

    def assertVariants(self, variants, targets):
        for i, j in zip(variants, targets):
            self.assertEqual(i.string, j)

    def test_novariant(self):
        variants = vc.Variant([], enzyme_obj)
        testres = vco._allocate_variants_to_peptide(variants, self.peptide)
        self.assertEqual(testres.enzyme_variants, [])
        self.assertEqual(testres.nonenzyme_variants, [])

    def test_variants(self):
        hgvs = [generatevariant(self.peptide, 13, 'W'),
                generatevariant(self.peptide, 20, 'K'),
                generatevariant(self.peptide, 25, 'G')
                ]
        variants = vc.Variant(hgvs, enzyme_obj)
        testres = vco._allocate_variants_to_peptide(variants, self.peptide)
        self.assertVariants(testres.enzyme_variants, ['p.Phe14Trp'])
        self.assertVariants(testres.nonenzyme_variants, ['p.Ser21Lys', 'p.Leu26Gly'])


class TestCleaveBreakerPeptide(unittest.TestCase):

    def test_cleavage_removed(self):
        peptide = vc.Peptide("ADPSGHCNFFWPSLFSHYKALM", 'identifier', 'gene')
        hgvs = [generatevariant(peptide, 10, 'K')]
        variants = vc.Variant(hgvs, enzyme_obj)
        peptide = vco._allocate_variants_to_peptide(variants, peptide)
        cleaved_within = vco._cleave_breaker_peptides(peptide)
        self.assertEqual([str(i) for i in cleaved_within[0]], ['ADPSGHCNFFW', 'PSLFSHYKALM'])
        self.assertEqual(str(cleaved_within[1][0]), 'ADPSGHCNFFKPSLFSHYKALM')

    def test_cleavage_added(self):
        peptide = vc.Peptide("ADPSGHCNFFKPSLFSHYKALM", 'identifier', 'gene')
        hgvs = [generatevariant(peptide, 10, 'W')]
        variants = vc.Variant(hgvs, enzyme_obj)
        peptide = vco._allocate_variants_to_peptide(variants, peptide)
        cleaved_within = vco._cleave_breaker_peptides(peptide)
        self.assertEqual(str(cleaved_within[0][0]), 'ADPSGHCNFFKPSLFSHYKALM')
        self.assertEqual([str(i) for i in cleaved_within[1]], ['ADPSGHCNFFW', 'PSLFSHYKALM'])


class TestDeduplicate(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.peptide = vc.Peptide("AAAAAAAWAAAAAAAAAAAAAAAAAAAAAAAAAAAAA", 'identifier', 'gene')

    def test_deduplicate(self):
        hgvs = [generatevariant(self.peptide, 7, 'K'),
                generatevariant(self.peptide, 22, 'W')]
        variants = vc.Variant(hgvs, enzyme_obj)
        peptide = vco._allocate_variants_to_peptide(variants, self.peptide)
        cleaved_within = vco._cleave_breaker_peptides(peptide)
        cleaved_peptides_within_length = [j for i in cleaved_within for j in i if j]
        cleaved_list = vco._deduplicate(cleaved_peptides_within_length)
        self.assertEqual(len(cleaved_list), 5)
        self.assertEqual(len(cleaved_list[4].applied_enzymevariants), 0)

class TestSprinkleVariants(unittest.TestCase):

    def assertsequenceprop(self, peptide, targets):
        for i, j in zip(peptide, targets):
            self.assertEqual(str(i), j)


    def test_sprinkle(self):
        peptide = vc.Peptide("FNGIYADPSGH", 'identifier', 'gene', start=20)
        hgvs = [generatevariant(peptide, 5, 'K'),
                generatevariant(peptide, 7, 'Y')]
        variants = vc.Variant(hgvs, enzyme_obj)
        peptide = vco._allocate_variants_to_peptide(variants, peptide)
        peptide_variant = vco._sprinkle_variants(peptide=peptide)
        self.assertsequenceprop(peptide_variant, ['FNGIYADYSGH',
                                                  'FNGIYKDPSGH',
                                                  'FNGIYKDYSGH'])
        variant_list = [j.string for i in peptide_variant for j in i.applied_nonenzymevariants]
        self.assertEqual(variant_list, ['p.Pro28Tyr', 'p.Ala26Lys', 'p.Ala26Lys', 'p.Pro28Tyr'])


class TestMiscleaveWithinBreakers(unittest.TestCase):

    default_max_length = vc.Peptide.max_length
    default_min_length = vc.Peptide.min_length

    @classmethod
    def setUpClass(cls):
        varpepdb.setpeptidelengths(6, 30)

    @classmethod
    def tearDownClass(cls):
        varpepdb.setpeptidelengths(cls.default_min_length, cls.default_max_length)

    def test_gap(self):
        peptide1 = vc.Peptide("FNGIYADPSG", 'identifier', 'gene', start=20, miscleave_count=0)
        peptide2 = vc.Peptide("FNGIYADPSG", 'identifier', 'gene', start=61, miscleave_count=0)
        peptide_list = [[peptide1, peptide2]]
        miscleaved_peptide = vco._miscleave_within_breakers(peptide_list)
        self.assertEqual(miscleaved_peptide, [])

    def test_nogap(self):
        peptide1 = vc.Peptide("FNGIYADPSG", 'identifier', 'gene', start=20)
        peptide2 = vc.Peptide("FNGIYADPSG", 'identifier', 'gene', start=30)
        peptide_list = [[peptide1, peptide2]]
        miscleaved_peptide = vco._miscleave_within_breakers(peptide_list)
        self.assertEqual(str(miscleaved_peptide[0]), 'FNGIYADPSGFNGIYADPSG')
        self.assertEqual(miscleaved_peptide[0].start, 20)
        self.assertEqual(miscleaved_peptide[0].end, 39)


class MiscleaveBetBreakers(unittest.TestCase):

    def test_2to1(self):
        peptide1 = vc.Peptide("FNGIYADPSG", 'identifier', 'gene', start=20, miscleave_count=0)
        peptide2 = vc.Peptide("DFNDFNDFND", 'identifier', 'gene', start=20, miscleave_count=0)
        peptide_list1 = [peptide1, peptide2]
        peptide3 = vc.Peptide("FNGIYADPSG", 'identifier', 'gene', start=32, miscleave_count=0)
        peptide4 = vc.Peptide("DFNDFNDFND", 'identifier', 'gene', start=31, miscleave_count=0)
        peptide_list2 = [peptide3, peptide4]
        miscl_pep_bet = vco._miscleave_bet_breakers(peptide_list1, peptide_list2)
        self.assertEqual(len(miscl_pep_bet), 0)

    def test_miscleave(self):
        peptide1 = vc.Peptide("FNGIYADPSG", 'identifier', 'gene', start=20, miscleave_count=0)
        peptide2 = vc.Peptide("DFNDFNDFND", 'identifier', 'gene', start=20, miscleave_count=0)
        peptide_list1 = [peptide1, peptide2]
        peptide3 = vc.Peptide("FNGIYADPSG", 'identifier', 'gene', start=30, miscleave_count=0)
        peptide4 = vc.Peptide("DFNDFNDFND", 'identifier', 'gene', start=31, miscleave_count=0)
        peptide_list2 = [peptide3, peptide4]
        miscl_pep_bet = vco._miscleave_bet_breakers(peptide_list1, peptide_list2)
        self.assertEqual(str(miscl_pep_bet[0]), 'FNGIYADPSGFNGIYADPSG')
        self.assertEqual(str(miscl_pep_bet[1]), 'DFNDFNDFNDFNGIYADPSG')

class TestGenerateSingle(unittest.TestCase):

    def assertsequenceprop(self, peptide, targets):
        for i, j in zip(peptide, targets):
            self.assertEqual(str(i), j)


    def test_generatesingle(self):
        aa_string = "FNGIYADPSGHCNFFWPSLFSHYKALMPAGNCI"
        peptide = vc.Peptide(aa_string, 'identifier', 'gene')
        res = varpepdb.generate_single(variants=[generatevariant(peptide, 3, 'A'), generatevariant(peptide, 20, 'W')],
                                       sequence=aa_string, gene='protein')
        self.assertsequenceprop(res, ['FNGAYADPSGHCNFFW', 
                                      'HYKALMPAGNCI', 
                                      'PSLFWHYKALMPAGNCI', 
                                      'FNGIYADPSGHCNFFWPSLFW', 
                                      'FNGAYADPSGHCNFFWPSLFW'])
        
    def test_nocanonicalmiscleavage(self):
        """Test that 2 canonical peptides without variants will not
            get miscleaved together
        """        
        aa_string = "FNGIYADWSGHCNFFWPSLFSHYKALMPAGNCI"
        peptide = vc.Peptide(aa_string, 'identifier', 'gene')
        res = varpepdb.generate_single(variants=[generatevariant(peptide, 32, 'K')],
                                        sequence=aa_string, gene='protein')
        
        self.assertsequenceprop(res, ['PSLFSHYKALMPAGNCK', 'SGHCNFFWPSLFSHYKALMPAGNCK'])

if __name__ == "__main__":
    unittest.main()



import itertools
import copy
import varpepdb.classes as vc
import varpepdb.utils as utils
from collections import Counter
import re
import multiprocessing as mp
from typing import List, Tuple, Union

miscleave = True
discardcanonical = True
cleaver = vc.Cleaver()
createcombi = True  # To create combinations of SAPs or not
message = True


def generate(input_list: List[Tuple[List[str], str, str, str]], proc: int = mp.cpu_count()) -> \
    List[vc.Peptide]:
    """Generates variant peptides with parallel processing

    Performs a check on the input list and then evokes ``generate_single`` for each input in parallel.

    Args:
        input_list: List of input. Each input element is a list consisting of the following 3 elements.                    
                    1. An list of HGVS nomenclatures describing amino acid substitution.
                        eg ['p.Gly1197Arg', 'p.Thr988Met',...]
                    2. Sequence of amino acid for the protein
                        'MTRSPPLRELPPSYTPPARTAAPQILAGSLKAPL...'
                    3. Gene name of the protein.
                        eg 'PTCH2'
                    4. Identifier for the protein
                        eg Uniprot accension number
        proc: Number of processes to run in parallel for each input.
                    Defaults to the number of cores available

    Raises:
        ValueError: If there are errors in the list of HGVS nomenclatures

    Returns:
        List of all variant peptides containing different combinations of the single amino acid substitutions.
    """

    # Check all inputs and show all errors
    error_list = []
    for idx in range(len(input_list)):

        hgvs_str_list = input_list[idx][0]

        correct_count = sum([bool(re.match(r'^p\.[A-Z][a-z]{2}\d+[A-Z][a-z]{2}', i)) for i in hgvs_str_list])
        if correct_count != len(hgvs_str_list):
            error_list.append(f'Element {idx} of input: One or more entries of variant list \
                              do not obey HGVS nomenclature.')
            
        if len(hgvs_str_list) == 0:
            if message:
                print('Variant list is empty. Generating only canonical peptides.')

        elif max(Counter(hgvs_str_list).values()) > 1:

            duplicated_list = []
            for item in Counter(hgvs_str_list).items():
                if item[1] > 1:
                    duplicated_list.append(item[0])

            error_list.append(f'Element {idx} of input: {", ".join(duplicated_list)} duplicated in variant list')

    if len(error_list) > 0:
        raise ValueError('\n'.join(error_list))

    # Run parallel processes
    with mp.Pool(proc) as pool:

        res = pool.starmap(generate_single, input_list)

    flatten_res = [j for i in res for j in i]

    return flatten_res


def generate_single(variants: List[str], sequence: str, gene: str, identifier: str) -> List[vc.Peptide]:
    """Generates variant peptides for one input.

    Args:
        variants: An list of HGVS nomenclatures describing amino acid substitution. eg. ['p.Gly1197Arg', ...]
        AA_wt_str: Sequence of amino acid for the protein.
        gene: Gene name of the protein. This is included as part of the description in the fasta header.

    Returns:
        List of all variant peptides containing different combinations of single amino acid substitutions.
    """
    # TODO: Include checking function    
    variants = vc.Variant(variants)
    peptide = vc.Peptide(sequence, identifier, gene)

    # Generate breaker peptides
    peptide_list = _generate_breaker_peptides(variants, peptide)

    cleaved_within_list = []
    cleaved_within_vars = []
    miscleaved_within_vars = []

    for peptide in peptide_list:

        peptide = _allocate_variants_to_peptide(variants, peptide)

        # List of peptides cleaved within breakers that are not too long
        cleaved_within = _cleave_breaker_peptides(peptide)

        # Select cleaved AA_sequence based on shortest variant list.
        # For peptides with more than 1 shortest variant list, add up the list.
        # Such peptides can be generated with more than 1 combination.
        cleaved_within_flat = [j for i in cleaved_within for j in i]
        cleaved_list = _deduplicate(cleaved_within_flat)
        cleaved_within_list.append(cleaved_list)

        # Sprinkle variants to cleaved peptides
        for cleaved_peptide in cleaved_list:

            cleaved_variant = _sprinkle_variants(peptide=cleaved_peptide)
            cleaved_within_vars.extend(cleaved_variant)

        if miscleave:

            # Miscleavaging within breakers
            miscl_pep_within = _miscleave_within_breakers(cleaved_within)
            miscleaved_list = _deduplicate(miscl_pep_within)

            # Sprinkle variants to peptides miscleaved within breakers
            for miscleaved_peptide in miscleaved_list:

                miscleaved_variant = _sprinkle_variants(peptide=miscleaved_peptide)
                miscleaved_within_vars.extend(miscleaved_variant)

    # Miscleavaging between breakers
    miscleaved_bet_vars = []
    if miscleave:

        for i in range(len(cleaved_within_list)-1):
            curr_list = cleaved_within_list[i]
            next_list = cleaved_within_list[i+1]

            if len(curr_list) == 0 or len(next_list) == 0:
                continue

            miscl_pep_bet = _miscleave_bet_breakers(curr_list, next_list)

            for pep in miscl_pep_bet:
                miscleaved_variant = _sprinkle_variants(peptide=pep)
                miscleaved_bet_vars.extend(miscleaved_variant)
    if message:
        print(f'completed {identifier}')

    return cleaved_within_vars + miscleaved_within_vars + miscleaved_bet_vars


def variant_containing_peptides(peptide_list: List[vc.Peptide]) -> List[vc.Peptide]:
    """Removes variant peptides that do not contain at least 1 amino acid substitution

    A segment of a protein may contain only a single amino acid substitution that causes enzymatic cleavage.
    One of the resulting peptides is a variant peptide that may not contain any amino acid substitution.
    For example, a p.Gly11Arg substitution in this protein FNGIYADPSGHNGYDA will result in FNGIYADPSR
    and HNGYDA after digestion by trypsin. HNGYDA is a variant peptide as it cannot be found by digesting
    the canonical sequence, but it doesn't contain any amino acid substitution within itself.

    This function removes such variant peptides that don't contain amino acid substitution.

    Args:
        peptide_list

    Returns:
        List of peptides containing at least 1 amino acid substitutions
    """

    informative_peptides = []

    for one_peptide in peptide_list:

        peptide_posrange = range(one_peptide.start, one_peptide.end + 1)
        allvariants = one_peptide.applied_nonenzymevariants + one_peptide.applied_enzymevariants

        # Keep canonical variants
        if not discardcanonical and len(allvariants) == 0:
            informative_peptides.append(one_peptide)
            continue

        for variant in allvariants:
            if variant.pos in peptide_posrange:

                informative_peptides.append(one_peptide)
                break

    return informative_peptides


def write(write_path: str, peptides: List[vc.Peptide], include_non_unique: bool = True) -> None:
    """Writes variant peptides into fasta file

    Example of an entry written into fasta file:
    >A0A8I5KQE6-v1.1 RPSA2 129-143 (p.Pro143Arg),p.Thr135Met 0
    ADHQPLMEASYVNLR

    'A0A8I5KQE6' is the sequence identifier of the parent protein (in this case the Uniprot ascension number) with
    'v1.1' is variant 1 of this peptide. 'v1.0' is the canonical peptide.    
    'RPSA2' is the name of the gene for this protein.
    '129-143' is the position of the parent protein sequence from which this peptide is dervied.
    '(p.Pro143Arg)' is an amino acid substitution that affected the enzyme cleavage site. Amino acid substitutions that
    introduce or remove cleavage sites are marked by parenthesis.
    'p.Thr135Met' is an amino acid subtstitution that didn't affect enzyne cleavage site.
    '0' refers to the number of miscleavages

    Args:
        write_path: Path of fasta file output
        peptide_list: List of variant peptide objects
        include_non_unique: Include variant peptides with non-unique sequences. Defaults to True. Identifier and
        description in the headers from each non-unique sequence are concatenated with a '/' as separator.

    Returns:
        None
    """

    unique_peptides, nonunique_peptides = separate_nonunique(peptides)

    unique_peptides.sort(key=lambda x: (x.identifier, x.start))

    fh = open(write_path, "w")

    v_group_counter = 0
    for _, peptide_group in itertools.groupby(unique_peptides, key=lambda x: (x.identifier, x.start, x.end)):

        v_counter = 0
        v_group_counter += 1

        for peptide in peptide_group:

            n_applied_enzymes = peptide.n_applied_enzymevar + peptide.n_applied_nonenzymevar

            # If canonical, use 0
            if n_applied_enzymes == 0:                        
                v_counter_actual = 0
            else:
                v_counter += 1
                v_counter_actual = v_counter
            identifier = peptide.identifier + '-v' + str(v_group_counter) + '.' + str(v_counter_actual)
            description = _make_description(peptide)

            # seqrecord = SeqRecord(Seq(str(peptide)),
            #                       id=identifier,
            #                       name=identifier,
            #                       description=description)

            utils.write_fasta(str(peptide), identifier, description, fh)

    if include_non_unique:

        name_tracker = Counter()

        for nonunique_list in nonunique_peptides:

            combined_identifier = '/'.join([i.identifier for i in nonunique_list])
            combined_description = '/'.join([_make_description(i) for i in nonunique_list])

            v_no = name_tracker[combined_identifier] + 1
            appended_identifier = combined_identifier + f'-v{v_no}'

            utils.write_fasta(str(nonunique_list[0]), appended_identifier, combined_description, fh)

            name_tracker[combined_identifier] += 1

    fh.close()


def separate_nonunique(peptide_list: List[vc.Peptide]) -> Tuple[List[vc.Peptide], List[List[vc.Peptide]]]:
    """Identifies variant peptide with unique and non-unique sequences

    Args:
        peptide_list: List of variant peptide objects.

    Returns:
        Tuple of 2. First element is a list of variant peptide with unique sequences. Second element is
        list of list of variant peptides sharing the same non-unique sequences.
    """
    peptide_list.sort(key=str)

    unique_seq = []
    duplicated_seq = []

    for _, peptide in itertools.groupby(peptide_list, key=str):
        pep_list = list(peptide)

        if len(pep_list) == 1:
            unique_seq.append(pep_list[0])

        else:
            duplicated_seq.append(pep_list)

    return unique_seq, duplicated_seq


def _get_tuple_combinations(variants: List[vc.SAP], omit: List[vc.SAP] = None) \
    -> List[Tuple[Tuple[int, str, Union[None, vc.SAP]]]]:
    """Creates all combinations of amino acid substitutions from a list of variants

    Args:
        variants: Variant object
        omit: Variants sharing the same positions found in this list are ignored.
        Defaults to None.

    Returns:
        List of tuples. Each tuple is a combination of amino acid substitutions to generate
        one variant peptide. An amino acid substitution is itself represented by a tuple. First element is position for
        substitution, second element is the amino acid to substitute in, third element is
        the variant object or ``None`` if this combination substitutes in the canonical amino
        acid.
    """

    if omit is None:
        omit = []
    enzyme_variant_pos = [i.pos for i in omit]
    variants[:] = [i for i in variants if i.pos not in enzyme_variant_pos]
    variants.sort(key=lambda x: x.pos)

    allpos_tuple_list = []

    for pos, sap in itertools.groupby(variants, key=lambda x: x.pos):
        sap_list = list(sap)
        onepos_tuple_list = []  # (position, AA to change to, SAP object)        
        if createcombi:
            # Add in reference amino acid
            onepos_tuple_list.append((pos, sap_list[0].ref, None))
        for sap_i in sap_list:
            onepos_tuple_list.append((pos, sap_i.alt, sap_i))
        allpos_tuple_list.append(onepos_tuple_list)
    tuple_combinations = list(itertools.product(*allpos_tuple_list))

    return tuple_combinations


def _allocate_variants_to_peptide(variants: vc.Variant, peptide: vc.Peptide) -> vc.Peptide:

    peptide = copy.deepcopy(peptide)
    for vartype in ('enzyme', 'nonenzyme'):
        for variant_i in getattr(variants, f'{vartype}_variants'):
            if variant_i in peptide:
                getattr(peptide, f'add_{vartype}_variants')(variant_i)

    return peptide


def _sprinkle_variants(peptide: vc.Peptide) -> List[vc.Peptide]:
    """Applies different combinations of amino acid substitutions

    Creates all possible combinations of amino acid substitutions that do not affect
    enzyme cleavage sites and applies each of them to generate variant peptides.
    If discardcanonical is True, canonical sequences will be discarded.

    Args:
        peptide: Peptide object

    Returns:
       List of variant peptides
    """

    peptide_variant = []

    # Omit nonenzyme variants acting on the same positions as applied enzyme variants to prevent the following situation:
    # Variant introducing a cleavage site (i.e. enzymevariant) is applied and resulted in cleavage
    # Another variant at that same position which doesn't cause cleavage overwrites the enzyme variant
    # Ends up with a cleaved peptide that doesn't have the cleavage site
    tuplecombi = _get_tuple_combinations(peptide.nonenzyme_variants, omit=peptide.applied_enzymevariants)

    for one_combi in tuplecombi:
        peptide_cp = copy.deepcopy(peptide)
        peptide_cp.apply_tuple_combination(one_combi, 'nonenzyme')

        # Exclude peptide if it is too short
        if not peptide_cp.within_length():
            continue

        # Exclude peptide if it contains 0 variant applied
        if discardcanonical:
            if peptide_cp.n_applied_enzymevar + peptide_cp.n_applied_nonenzymevar == 0:
                continue
        peptide_variant.append(peptide_cp)

    return peptide_variant


def _deduplicate(cleaved_peptides_flatten: List[vc.Peptide]) -> List[vc.Peptide]:
    """Deduplicate cleaved peptides

    Different combinations of amino acid substitutions may produce the same peptide after
    enzymatic cleaving. This function selects for the most parsimonious combination.

    Args:
        cleaved_peptides_flatten: List of peptide after cleaving

    Returns:
        List of peptide with unique sequences and most straightforward amino acid substitutions.
    """
    cleaved_peptides_flatten_c = copy.deepcopy(cleaved_peptides_flatten)
    # Group by seq, and start/end position. Positions are needed to differentiate repeated motifs.
    cleaved_peptides_flatten_c.sort(key=lambda x:(str(x), x.start, x.end))    
    group_iter = itertools.groupby(cleaved_peptides_flatten_c, key= lambda x:(str(x), x.start, x.end))

    dedup_list = []
    for _, peptide_group in group_iter:
        peptide_list = list(peptide_group)

        min_nvariants = min([i.n_applied_enzymevar for i in peptide_list])

        min_idx_list = []
        for idx, pep in enumerate(peptide_list):
            if pep.n_applied_enzymevar == min_nvariants:
                min_idx_list.append(idx)

        firstpeptide = peptide_list[min_idx_list[0]]

        # If there is more than 1 combination of enzymatic SAPs leading to the same cleaved peptide, just use the first combination found
        # and print a warning
        if len(min_idx_list) > 1:
            if message:
                print('Warning: More than 1 combination of enymatic SAPs producing the same peptide found')

        dedup_list.append(firstpeptide)

    return dedup_list


def _miscleave_within_breakers(cleaved_within: List[List[vc.Peptide]]) -> List[vc.Peptide]:

    miscleaved_peptide_list = []
    # one_combi_cleaved is list of cleaved peptides after applying 1 tuple combination
    for one_combi_cleaved in cleaved_within:
        # Apply miscleavages within each one_combi_cleaved
        for i in range(len(one_combi_cleaved)-1):
            if one_combi_cleaved[i+1].start - one_combi_cleaved[i].end == 1:

                miscleaved_peptide = one_combi_cleaved[i] + one_combi_cleaved[i+1]

                if miscleaved_peptide.within_length():
                    miscleaved_peptide_list.append(miscleaved_peptide)

            elif one_combi_cleaved[i+1].start - one_combi_cleaved[i].end - 1 <= vc.Peptide.max_length:
                raise RuntimeError(f'Peptide within length limits missing in generating miscleaved peptides for \
                                   {one_combi_cleaved[i].identifier}')

            else:
                # Contiguous peptide was longer than limit and removed. Miscleave peptide would have been too long.
                pass

    return miscleaved_peptide_list


def _cleave_breaker_peptides(peptide: vc.Peptide) -> List[vc.Peptide]:

    # Generate tuple combinations
    tuple_combinations = _get_tuple_combinations(peptide.enzyme_variants)

    # Generate cleaved AA_Sequence for all combinations of enzyme SAPs within breakers

    if len(tuple_combinations) == 0:
        cleave_result = cleaver.cleave(peptide)

        if cleave_result.nb_cleavage > 0:
            raise RuntimeError('Cleavage occurred within breaker without enzyme variants!')

        if peptide.not_too_long():  # Keep peptides for miscleavaging later
            peptide.miscleave_count = 0
            cleaved_peptides = [peptide]
        else:
            cleaved_peptides = []

    else:
        # Apply variants and cleave
        cleaved_peptides = []
        for tuple_combinations_i in tuple_combinations:
            peptide_cp = copy.deepcopy(peptide)
            peptide_cp.apply_tuple_combination(tuple_combinations_i, variant_type='enzyme')
            applied_cleaved = cleaver(peptide_cp)
            if not createcombi:
                # If not creating combi, replace certain peptides with canonical peptides
                # To account for certain edge cases like
                # SEGGTAQLLRR with a subsitution at the C terminus. 
                # SEGGTAQLLR could be generated canonically without any enzyme-SAP
                # So should be considered as canonical
                non_applied_cleaved = cleaver(peptide)
                for i in range(len(applied_cleaved)):
                    for j in non_applied_cleaved:
                        if applied_cleaved[i] == j:
                            applied_cleaved[i] = j
                            
            cleaved_peptides.append(applied_cleaved)

    return cleaved_peptides


def _generate_breaker_peptides(variants: vc.Variant, peptide: vc.Peptide) -> List[vc.Peptide]:

    WT_cleave_result = cleaver.cleave(peptide)
    positions_to_remove = []

    for enzyme_variant in variants.enzyme_variants:
        positions_to_remove.extend(range(enzyme_variant.pos + variants.upstream,
                                         enzyme_variant.pos + variants.downstream + 1))

    cleave_pos_str = WT_cleave_result.get_cleavage_pos()

    if len(cleave_pos_str) > 0:
        cleave_pos = set([int(i) for i in WT_cleave_result.get_cleavage_pos().split(',')])
    else:  # No endogenous cleave sites
        cleave_pos = set()

    cleave_pos_leftover = [0] + sorted(list(cleave_pos.difference(positions_to_remove))) + [len(peptide)]

    peptide_list = []
    for ind in range(len(cleave_pos_leftover)-1):
        peptide_list.append(peptide[cleave_pos_leftover[ind]:cleave_pos_leftover[ind+1]])
    return peptide_list


def _miscleave_bet_breakers(curr_list: List[vc.Peptide], next_list: List[vc.Peptide]) -> List[vc.Peptide]:

    miscleaved_bet = []

    curr_end_pos = max([i.end for i in curr_list])
    next_start_pos = min([i.start for i in next_list])

    if next_start_pos == curr_end_pos + 1:

        curr_idx = []
        for idx, peptide in enumerate(curr_list):
            if peptide.end == curr_end_pos:
                curr_idx.append(idx)

        next_idx = []
        for idx, peptide in enumerate(next_list):
            if peptide.start == next_start_pos:
                next_idx.append(idx)

        for i, j in itertools.product(curr_idx, next_idx):

            miscleaved_peptide = curr_list[i] + next_list[j]

            if miscleaved_peptide.within_length():
                miscleaved_bet.append(miscleaved_peptide)

    return miscleaved_bet


def _make_description(peptide: vc.Peptide) -> str:
    """Makes description for fasta entry header

    Args:
        peptide: Variant peptide object

    Raises:
        RuntimeError: If peptide doesn't arise from an amino acid substitution and can be
        obtained from the canonical sequence

    Returns:
        Description for fasta entry header
    """

    enzymevar = ','.join([i.string for i in peptide.applied_enzymevariants])
    nonenzymevar = ','.join([i.string for i in peptide.applied_nonenzymevariants])

    if peptide.applied_enzymevariants and peptide.applied_nonenzymevariants:
        variant_string = '(' + enzymevar + '),' + nonenzymevar

    elif peptide.applied_enzymevariants and not peptide.applied_nonenzymevariants:
        variant_string = '(' + enzymevar + ')'

    elif not peptide.applied_enzymevariants and peptide.applied_nonenzymevariants:
        variant_string = nonenzymevar

    else:
        variant_string = 'Canonical'

    pos_range = str(peptide.start + 1) + '-' + str(peptide.end + 1)

    description = ' '.join([peptide.gene, pos_range, variant_string, str(peptide.miscleave_count)])

    return description



from rpg.sequence import Sequence as Rpg_Sequence
import varpepdb.utils as utils
import rpg
import functools
import numbers
from typing import Tuple, Union


@functools.total_ordering
class SAP:
    def __init__(self, pos, ref, alt, string):
        self.pos = pos
        self.ref = ref
        self.alt = alt
        self.string = string

    def __lt__(self, other):
        return self.pos < other

    def __eq__(self, other):
        return self.pos == other

    def __repr__(self):
        return f"(pos={self.pos}, ref={self.ref}, alt={self.alt}, string={self.string})"


class Variant:

    enzyme = None

    def __init__(self, hgvs_str_list=None, AC=None):
        self.variants = []

        if hgvs_str_list is not None:
            self.parse_list(hgvs_str_list)

        self._get_cleave_position()
        self._get_flank_dist()
        self._get_enzyme_variants()
        self._get_nonenzyme_variants()

    def __len__(self):
        return len(self.variants)

    def __repr__(self):
        return repr(self.variants)

    def _sort(self):
        self.variants.sort(key=lambda x: x.pos)

    def _get_rule_attrib(self, rulelist, attrib):
        if len(rulelist) == 0:
            return []

        elif len(rulelist) == 1:
            return [getattr(rulelist[0], attrib)] + self._get_rule_attrib(rulelist[0].rules, attrib)

        elif len(rulelist) > 1:
            return [getattr(rulelist[0], attrib)] + self._get_rule_attrib(rulelist[0].rules, attrib) + \
                self._get_rule_attrib(rulelist[1:], attrib)

    def _get_cleave_position(self):
        position_list = []

        for enzyme_i in self.enzyme:
            position_list.extend(self._get_rule_attrib(enzyme_i.rules, 'pos'))

        self.position_list = set(position_list)

    def _get_flank_dist(self):
        index_list = []

        for enzyme_i in self.enzyme:
            index_list.extend(self._get_rule_attrib(enzyme_i.rules, 'index'))
        max_downstream = max(index_list)
        min_upstream = min(index_list)

        enzyme_recog_dist = max_downstream - min_upstream
        if enzyme_recog_dist == 0:
            self.downstream = 0
            self.upstream = 0

            if 1 in self.position_list:
                self.downstream = 1

            if 0 in self.position_list:
                self.upstream = -1

        elif enzyme_recog_dist > 0:
            self.downstream = enzyme_recog_dist
            self.upstream = -enzyme_recog_dist

        elif enzyme_recog_dist < 0:
            raise Exception(f"enzyme_recog_dist is {enzyme_recog_dist}!")

    def _get_enzyme_variants(self):
        AA_list = []

        for enzyme_i in self.enzyme:
            AA_list.extend(self._get_rule_attrib(enzyme_i.rules, 'amino_acid'))

        self.enzyme_AAs = set(AA_list)
        self.enzyme_variants = []

        for i in self.variants:
            if i.ref in self.enzyme_AAs or i.alt in self.enzyme_AAs:
                self.enzyme_variants.append(i)

    def _get_nonenzyme_variants(self):
        self.nonenzyme_variants = []

        for i in self.variants:
            if i.ref not in self.enzyme_AAs and i.alt not in self.enzyme_AAs:
                self.nonenzyme_variants.append(i)

    def parse_list(self, hgvs_str_list):
        for variant_one in hgvs_str_list:
            
            pos = int(variant_one[5:-3]) - 1  # Use 0 based position
            ref = utils.AA_3to1[variant_one[2:5]]
            alt = utils.AA_3to1[variant_one[-3:]]

            self.variants.append(SAP(pos, ref, alt, string=variant_one))

        self._sort()


class Peptide:
    max_length = 30
    min_length = 6

    def __init__(self, aa_string, identifier, gene, start=0, enzyme_variants=None, nonenzyme_variants=None,
                 miscleave_count=0, applied_enzymevariants=None, applied_nonenzymevariants=None):

        if isinstance(aa_string, str):
            self._seq = bytearray(aa_string, "ASCII")
        elif isinstance(aa_string, bytearray):
            self._seq = aa_string
        else:
            raise TypeError(
                "aa_string should be a string or bytearray object"
            )

        self.ori = self._seq
        self.identifier = identifier
        self.gene = gene
        self.start = start  # 0 based

        if enzyme_variants is None:
            enzyme_variants = []
        self.enzyme_variants = enzyme_variants

        if nonenzyme_variants is None:
            nonenzyme_variants = []
        self.nonenzyme_variants = nonenzyme_variants

        self.miscleave_count = miscleave_count

        if applied_enzymevariants is None:
            applied_enzymevariants = []
        self.applied_enzymevariants = applied_enzymevariants

        if applied_nonenzymevariants is None:
            applied_nonenzymevariants = []
        self.applied_nonenzymevariants = applied_nonenzymevariants

        self.n_applied_enzymevar = len(applied_enzymevariants)
        self.n_applied_nonenzymevar = len(applied_nonenzymevariants)

    def __eq__(self, other):
        # Compare only positions in protein and sequence
        return self.seq() == other.seq() and self.start == other.start and self.end == other.end and self.identifier == other.identifier

    def __contains__(self, other: SAP):
        return other.pos >= self.start and other.pos <= self.end

    def __add__(self, other: 'Peptide'):
        """Concentenate 2 Peptide objects
        """
        if self.end != other.start-1:
            raise RuntimeError('Sequences not contiguous, cannot be added.')

        if self.identifier != other.identifier:
            raise RuntimeError('Different sequence identifiers, cannot be added.')

        newenzyme_variants = self.enzyme_variants[:] + \
            [i for i in other.enzyme_variants if i not in self.enzyme_variants]

        newnonenzyme_variants = self.nonenzyme_variants + other.nonenzyme_variants

        newmiscleave_count = self.miscleave_count + other.miscleave_count + 1

        new_applied_enzymevariants = self.applied_enzymevariants[:] + \
            [i for i in other.applied_enzymevariants if i not in self.applied_enzymevariants]
        new_applied_nonenzymevariants = self.applied_nonenzymevariants + other.applied_nonenzymevariants

        newinstance = self.__class__(aa_string=self._seq+other._seq,
                                     identifier=self.identifier,
                                     gene=self.gene,
                                     start=self.start,
                                     enzyme_variants=newenzyme_variants,
                                     nonenzyme_variants=newnonenzyme_variants,
                                     miscleave_count=newmiscleave_count,
                                     applied_enzymevariants=new_applied_enzymevariants,
                                     applied_nonenzymevariants=new_applied_nonenzymevariants)
        return newinstance

    def __str__(self):
        return self.seq()
    
    def seq(self):
        return self._seq.decode("ASCII")

    def __repr__(self):
        return self.seq() + '\n' + f'Start: {self.start}' + '\n' + f'End: {self.end}'

    def __getitem__(self, index):

        if isinstance(index, numbers.Integral):
            start = index + self.start
            end = index + self.start
            newseq = chr(self._seq[index])
        else:
            start = self.start + (index.start or 0)
            end = self.start + index.stop - 1  # Slice doesn't include last element
            newseq = self._seq[index]
        # Todo: negative slice

        new_nonenzyme_variants = [i for i in self.nonenzyme_variants if i.pos in range(start, end+1)]
        new_applied_nonenzymevariants = [i for i in self.applied_nonenzymevariants if i.pos in range(start, end+1)]

        return self.__class__(aa_string=newseq,
                              identifier=self.identifier,
                              gene=self.gene,
                              start=start,
                              enzyme_variants=self.enzyme_variants[:],
                              nonenzyme_variants=new_nonenzyme_variants,
                              miscleave_count=self.miscleave_count,
                              applied_enzymevariants=self.applied_enzymevariants[:],
                              applied_nonenzymevariants=new_applied_nonenzymevariants)

    def __setitem__(self, index, value: str):
        """Set a subsequence of single letter via value parameter.

        >>> my_seq = Peptide('ACTCGACGTCG')
        >>> my_seq[0] = 'T'
        >>> my_seq
        Peptide('TCTCGACGTCG')
        """
        if isinstance(index, numbers.Integral):
            # Replacing a single letter with a new string
            self._seq[index] = ord(value)
        else:
            # Replacing a sub-sequence
            if isinstance(value, Peptide):
                self._seq[index] = value._seq
            elif isinstance(value, str):
                self._seq[index] = value.encode("ASCII")
            else:
                raise TypeError(f"received unexpected type '{type(value).__name__}'")

    def __len__(self):
        return len(self._seq)

    @property
    def start(self):
        """Setting the start position automatically sets the end
        """
        return self._start

    @start.setter
    def start(self, value):
        self._start = value
        self._end = value + len(self._seq) - 1

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, value):
        raise AttributeError('End position is read-only. Set the start position only. ')

    def apply_tuple_combination(self, one_combination: Tuple[int, str, Union[None, SAP]], variant_type: str):
        """Apply one amino acid substitution

        Args:
            one_combination: A tuple describing the amino acid substition.
                May also substitute back to canonical sequence.
            variant_type: enzyme or nonenzyme
        """

        if variant_type not in ['enzyme', 'nonenzyme']:
            raise ValueError('variant_type must be enzyme or nonenzyme')

        for one_change in one_combination:
            skip_current_change = False
            sap_string = one_change[2].string if one_change[2] else f"ref {one_change[1]} at pos {one_change[0]}"

            for i in self.applied_nonenzymevariants:
                if one_change[0] == i.pos:
                    raise RuntimeError(f'{self.identifier}: {i.string} already applied. \
                                    Applying {sap_string} overwrites it.')

            for i in self.applied_enzymevariants:
                if one_change[0] == i.pos:
                    print(f'{self.identifier}: {i.string} enzyme variant already applied. Ignoring {sap_string}. ')
                    skip_current_change = True

            if skip_current_change:
                continue

            index = one_change[0]-self.start  # Convert global position to local position
            target = one_change[1]

            self[index] = target

            if one_change[2]:  # If there is a SAP object
                if variant_type == 'enzyme':
                    self.applied_enzymevariants.append(one_change[2])
                    self.n_applied_enzymevar += 1

                elif variant_type == 'nonenzyme':
                    self.applied_nonenzymevariants.append(one_change[2])
                    self.n_applied_nonenzymevar += 1

    def add_enzyme_variants(self, v):
        self.check_variant(v)
        self.enzyme_variants.append(v)

    def add_nonenzyme_variants(self, v):
        self.check_variant(v)
        self.nonenzyme_variants.append(v)

    def check_variant(self, v):
        if str(self[v.pos-self.start]) != v.ref:
            raise RuntimeError(f'Reference amino acid is wrong in {v.string} for {self.identifier}.')

    def convert_rpg(self):
        return Rpg_Sequence(self.identifier, str(self))

    def restore_original(self):
        self._seq = self.ori
        self.applied_nonenzymevariants = []

    def within_length(self):
        return len(self) >= self.min_length and len(self) <= self.max_length

    def not_too_long(self):
        return len(self) <= self.max_length


class Cleaver:
    """Cleave peptide object with enzyme

    Returns:
       List of cleaved peptides. Cleaved peptides that exceed length limit are excluded.
    """
    enzyme = None

    def __init__(self, aa_pka=rpg.core.AA_PKA_IPC, aa_mass=rpg.core.AA_MASS_AVERAGE, water_mass=rpg.core.WATER_MASS):

        self.aa_pka = aa_pka
        self.water_mass = water_mass
        self.aa_mass = aa_mass

    def __call__(self, aaseq: 'Peptide'):

        cleave_result = self.cleave(aaseq)
        cleave_positions = [0] + [int(i) for i in cleave_result.get_cleavage_pos().split(',') if i] + [len(aaseq)]
        cleaved_peptides_list = []

        for ind in range(len(cleave_positions)-1):
            cleaved_peptide = aaseq[cleave_positions[ind]:cleave_positions[ind+1]]

            if cleaved_peptide.not_too_long():
                cleaved_peptide.miscleave_count = 0
                cleaved_peptides_list.append(cleaved_peptide)
            else:
                continue

        return cleaved_peptides_list

    def __repr__(self):
        return f"Cleaver object with Enzyme(s) = {', '.join([i.name for i in Cleaver.enzyme])}"

    def cleave(self, aaseq: 'Peptide'):
        cleave_result = rpg.digest.digest_one_sequence(seq=aaseq.convert_rpg(),
                                                       enz=Cleaver.enzyme,
                                                       mode='concurrent',
                                                       aa_pka=self.aa_pka,
                                                       aa_mass=self.aa_mass,
                                                       water_mass=self.water_mass)[0]
        return cleave_result
